
struct coeffs_struct
  x::Vector{Float64}
  p::Vector{Float64}
  q::Vector{Float64}
  r::Vector{Float64}
end
struct diagonal_struct
  a1f::Vector{Float64}
  a2f::Vector{Float64}
  a3f::Vector{Float64}
end

struct BalloonParams
  k_w::Int64
  l_geom_input::Bool
  l_tokamak_input::Bool
  kth::Int64

  r0::Float64
  amin::Float64
  beta0::Float64
  b0_v::Float64
end  

"""
    cobra(vmec_data, vmec, surfs, nwells, θs, ζs, mode=1, geom_input=true, tokamak_input=false, lscreen=true)

Calculates ballooning equilibrium based off of the CobraVmec code. Originally converted from fortran to Julia by Sanket Patil, rewritten by Aaron Bader

# Arguments
 - `vmec_data::VMEC.VmecData`: a vmec_data file (will become obsolete at some point)
 - `vmec::VMEC.Vmec`: A Vmec struct
 - `slist::Vector{Int64}`: list of s values
 - `nwells::Int64`: Number of helical wells to include
# Optional inputs
 - `θs::Vector{Float64}`: Vector of starting θ values (in radians)
 - `ζs::Vecotr{Float64}`: Vector of starting ζ values (in radians normalized to field periods)
 - `mode::Int64`: mode number, 1 is the most unstable mode
 - `geom_input::Bool`: Flag, default true
 - `tokamak_input::Bool`: Flag, default false, not sure we'll ever use this
 - `lscreen::Bool`: Flag, output to screen, default true 

"""
function cobra(vmec::Vmec, slist::Vector{Float64}, nwells::Int64;
               θs::Vector{Float64} = [0.0, π/8, π/4, 2*π/8, π/2], 
               ζs::Vector{Float64} = [0.0, π/2, π, 2*π/3, 2*π],
               mode::Int64=1, geom_input::Bool= true, tokamak_input::Bool=false, lscreen::Bool = true)
  #=
  _____________________________________________________________________________________
  |                                                                                   |
  |              COBRA (COde for Ballooning Rapid Analysis)                           |
  |                     ==       =          =     =                                   |
  |                                                                                   |
  |      VERSION: julia rewrite of                                                    |
  |               4.1  (FORTRAN 90: fixed FORMAT; standard I/O; VMEC coordinates)     |
  |                    (Accepts ASYMMETRIC input from VMEC via LASYM)                 |
  |                    (Accepts initial line location as geometrical or (ALHPA,ANGLE) |
  |                    (Good for LRFP=T or F)                                         |
  |      Last update: 03/27/13                                                        |
  |                                                                                   |
  |      AUTHORS:   R. Sanchez                                                        |
  |             Universidad Carlos III de Madrid, Leganes 28911, SPAIN.               |
  |                 S.P. Hirshman                                                     |
  |             Oak Ridge National Laboratory, Oak Ridge, TN 37831-9701, USA          |
  |      JULIA REWRITE: S. Patil, UW-Madison                                          |                                                                             |
  |      REFERENCES:                                                                  |
  |                                                                                   |
  |        1."COBRA: and optimized code for fast analysis of ideal ballooning         |
  |        stability of 3-D magnetic equilibria", R. Sanchez, S.P. Hirshman, J.C.     |
  |        Whitson and A.S. Ware, submitted to Journal of Computational Physics (1999)|
  |                                                                                   |
  |        2."Improved magnetic coordinate representation for ideal ballooning        |
  |        stability calculations with the COBRA code", R. Sanchez, S.P. Hirshman     |
  |        H.V. Wong, to be submitted to Journal of Computational physics (2000)      |
  |                                                                                   |
  |      DISCLAIMER:  This code is under development by R.Sanchez at the Departamento |
  |        de Fisica, Universidad Carlos III de Madrid, SPAIN. As a BETA version, the |
  |        code is supplied on "as it is" basis, and the non-existence of "bugs" of   |
  |        ANY kind is NOT guaranteed. Any problem or comment should be reported to   |
  |        R. Sanchez at raul.sanchez@uc3m.es                                         |
  _____________________________________________________________________________________
  =#
  #this will still crash if surfaces are too small, can determine minimum value
  #but better to allow extrapolation
  slist = slist[slist .> 0.001]
  slist = slist[slist .< 0.999]
  ns = length(slist)
  vmec_data = VmecData(vmec)

  # Add here:  date and time if wanted
  #vmec, vmec_data = VMEC.readVmecWout(extension)
  (r0, amin, beta0, b0_v) = order_input(vmec_data);

  params = BalloonParams(nwells, geom_input, tokamak_input, mode, r0, amin, beta0, b0_v)
  println("wout file read successfully!")


  # Solve ballooning equation

  balloonGrowth = zeros(length(slist), length(θs), length(ζs))
  mercier = zeros(length(slist))
  pres = zeros(length(slist))
  xmax = zeros(length(slist), length(θs), length(ζs))
  np = zeros(length(slist), length(θs), length(ζs))
  tsym = zeros(Int8, length(slist), length(θs), length(ζs))
  nRichardson = zeros(length(slist), length(θs), length(ζs))
  μ0 = 4*π*1.0E-7
  for (i_s,s) in enumerate(slist)
    
    vmecSurface = VmecSurface(s, vmec);
    pres[i_s] = vmecSurface.pres[1]*μ0
    mercier = vmecSurface.DMerc[1]

    for (iθ, init_theta) in enumerate(θs)             # Loop over all poloidal, toroidal initial pairs
      for (iζ, init_zeta) in enumerate(ζs)
        #println("at: ",s," ",init_theta," ",init_zeta)

        # the following to be used as args in get_ballooning_grate
        init_zeta = mod(init_zeta, 2*π)/vmec.nfp;
        init_theta = mod(init_theta, 2*π);

        lfail_balloon = false 
        
        # get growth rates on desired surfaces
        bg, xm, npt, j, tsymm = get_ballooning_grate(vmecSurface, init_zeta, init_theta, params);
        balloonGrowth[i_s, iθ, iζ] = bg
        xmax[i_s, iθ, iζ] = xm
        np[i_s, iθ, iζ] = npt
        nRichardson[i_s, iθ, iζ] = j
        tsym[i_s, iθ, iζ] = tsymm
      end
    end
  end

  if lscreen
    for (iθ, θ) in enumerate(θs)
      for (iζ, ζ) in enumerate(ζs)
        println("-"^125)
        if params.l_geom_input
          @printf("   NS      FLUX-s        ZT_0        TH_0       GR. RATE    IT  POINTS        XMAX      OK?     SYMM        PRES         MERC\n")
        elseif params.l_tokamak_input
          @printf("   NS      FLUX-s       ZETAK    ALPHA_ST       GR. RATE    IT  POINTS        XMAX      OK?     SYMM        PRES         MERC\n")
        else
          @printf("   NS      FLUX-s      THETAK   ALPHA_TOK       GR. RATE    IT  POINTS        XMAX      OK?     SYMM        PRES         MERC\n")
        end
        println("-"^125)
        ζp = ζ*180/π / vmec.nfp
        θp = θ*180/π
        
        for (i_s, s) in enumerate(slist)
          surf = Int(round((s * (vmec.ns - 1)) + 1))
          j = nRichardson[i_s, iθ, iζ]
          bg = balloonGrowth[i_s, iθ, iζ]
          xm = xmax[i_s, iθ, iζ]
          npt = np[i_s, iθ, iζ]
          if bg == NaN
            l_success = "F"
          else
            l_success = "T"
          end
          tsymm = tsym[i_s, iθ, iζ]
          if params.l_geom_input
            @printf("%5i    %8.2e    %8.2e    %8.2e    %11.4e    %2i    %4i    %8.2e    %5s    %5s    %8.2e    %8.2e\n",
                    surf, s, ζp, θp, bg, j, npt, xm, l_success, tsymm, pres[i_s], mercier[1])
          elseif params.l_tokamak_input
            @printf("%5i    %8.2e    %8.2e    %8.2e    %11.4e    %2i    %4i    %8.2e    %5s    %5s    %8.2e    %8.2e\n",
                    surf, s, θp, ζp, bg, j, npt, xm, l_success, tsymm, pres[i_s], mercier[1])
          else
            @printf("%5i    %8.2e    %8.2e    %8.2e    %11.4e    %2i    %4i    %8.2e    %5s    %5s    %8.2e    %8.2e\n",
                    surf, s, ζp, θp, bg, j, npt, xm, l_success, tsymm, pres[i_s], mercier[1])
          end
        end

        if params.l_geom_input
           @printf("ZETA0 = %10.3e THETA0 = %10.3e\n", ζ, θ);
        elseif params.l_tokamak_input
            @printf("ALPHA_TOK = %10.3e THETAK = %10.3e\n", ζ, θ);
        else
            @printf("ALPHA_ST = %10.3e ZETAK = %10.3e\n", θ, ζ);
          end
        end
           @printf("____________________________________________________\n");


    end # zeta loop
  end # theta loop


end
