using .MPI

function gamma_c_target(surfaces::Union{Float64,Vector{Float64}}, vmec::Vmec, zetaMin::Float64, 
                        zetaStep::Float64, zetaMax::Float64,
                        BResolution::Int64, comm::MPI.Comm, epsEff::Bool=true, gammac::Bool=true) 
  
  mpiRank=MPI.Comm_rank(comm)
  nTargetValues = length(surfaces)
  if epsEff && gammac
    targetValues = Array{Tuple{Float64, Float64}}(undef, nTargetValues)
  elseif epsEff || gammac
    targetValues = Array{Float64,1}(undef,nTargetValues)
  else
    println("must select either epsEff or gammac")
  end
  #println(nTargetValues)

  for index = 1:nTargetValues
    vmecSurface = VmecSurface(surfaces[index],vmec)

    targetTemp = compute_gamma_c(vmecSurface,zetaMin, zetaStep, zetaMax, BResolution, comm,
                                 epsEff, gammac) 
    if mpiRank == 0
      targetValues[index] = targetTemp
    end
  end
  if mpiRank == 0
    return targetValues
  else
    return nothing
  end
end

function gamma_c_target(vmecSurface::VmecSurface, zetaMin::Float64, zetaStep::Float64, zetaMax::Float64,
                        BResolution::Int64, comm::MPI.Comm, epsEff::Bool=true, gammac::Bool=true)
  if epsEff && gammac
    targetValues = (Inf,Inf)
  elseif epsEff || gammac
    targetValues = Inf
  else
    println("must select either epsEff or gammac")
  end

  return compute_gamma_c(vmecSurface, zetaMin, zetaStep, zetaMax, BResolution, comm, epsEff, gammac) 
end

function compute_gamma_c(vmec::VmecSurface{T},
                         zetaMin::A,
                         zetaStep::A,
                         zetaMax::A,
                         BResolution::Int64,
                         comm:: MPI.Comm,
                         epsEff::Bool = true,
                         gammac::Bool = true;
                        ) where {T,A}

  #set up the needed MPI information
  mpiRank=MPI.Comm_rank(comm)
  mpiSize=MPI.Comm_size(comm)

  chunkBase = div(BResolution,mpiSize)
  chunkExtra = mod(BResolution,mpiSize)
  #construct the chunkSize and chunkStart arrays
  chunkStarts = Vector{Int64}(undef, mpiSize)
  chunkSizes = Vector{Int64}(undef, mpiSize)
  chunkStarts[1] = 1
  chunkSizes[1] = chunkBase
  if chunkExtra > 0
    chunkSizes[1] += 1
  end
  for i in 2:mpiSize
    chunkStarts[i] = chunkStarts[i- 1] + chunkSizes[i-1]
    chunkSizes[i] = i <= chunkExtra ? chunkBase + 1 : chunkBase
  end

  start = chunkStarts[mpiRank+1]
  chunkSize = chunkSizes[mpiRank+1]

  H2overI_b=Vector{Float64}(undef, BResolution)
  gammac_b=Vector{Float64}(undef, BResolution)
  bprime_b=Vector{Float64}(undef,BResolution)

  dsOverB, gradPsiOverB = compute_gamma_c!(H2overI_b,gammac_b,bprime_b,
                                           vmec,zetaMin,zetaStep,zetaMax,
                                           BResolution,start,start+chunkSize-1)

  #Gather all the results
  if mpiRank == 0
    MPI.Gatherv!(MPI.IN_PLACE, VBuffer(H2overI_b,chunkSizes), 0, comm)
    MPI.Gatherv!(MPI.IN_PLACE, VBuffer(gammac_b,chunkSizes), 0, comm)
    MPI.Gatherv!(MPI.IN_PLACE, VBuffer(bprime_b,chunkSizes), 0, comm)
  else
    MPI.Gatherv!(H2overI_b[start:start+chunkSize-1],nothing,0,comm)
    MPI.Gatherv!(gammac_b[start:start+chunkSize-1],nothing,0,comm)
    MPI.Gatherv!(bprime_b[start:start+chunkSize-1],nothing,0,comm)
  end

  #println("Time to compute integral targets: "*string(looptime))
  MPI.Barrier(comm)
  
  if mpiRank > 0
    return nothing
  end

  if epsEff 
    H2overI = H2overIIntegral(bprime_b, H2overI_b)
    EpsEff32 = H2overI/(gradPsiOverB)^2*dsOverB * pi * vmec.Rmajor_p^2/8/sqrt(2)
    #println("epsfactors: ",vmec.Rmajor_p," ",dsOverB," ",gradPsiOverB)
    #println(EpsEff32," ",EpsEff32^(2.0/3))
  end

  if gammac 
    gammacInt = gammacIntegral(bprime_b, gammac_b)
    bigGammac = pi / 2 / sqrt(2) * gammacInt / dsOverB
  end

  if epsEff && gammac 
    target = (EpsEff32^(2.0/3), bigGammac)
  elseif epsEff
    target = EpsEff32^(2.0/3)
  else
    target = bigGammac
  end

  return target
end


